﻿using C1.Win.C1FlexGrid;
using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Entity;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;

namespace C1FlexGridOutlinesAndTreesBound
{
    public partial class Form1 : Form
    {

        //------------------------------------------------------------------------------------
        #region ** fields

        CellStyle[] _nodeStyle;
        northwndEntities _ctx;
        List<OrderInfo> _ordersData;

        #endregion

        //------------------------------------------------------------------------------------
        #region ** ctor

        public Form1()
        {
            InitializeComponent();

            // get data
            _ctx = new northwndEntities();

            DbSet<Order> orders = _ctx.Orders;
            DbSet<Order_Detail> ordersDetails = _ctx.Order_Details;

            IQueryable<OrderInfo> ordersQuery = from order in orders
                                                join orderDetail in ordersDetails on order.OrderID equals orderDetail.OrderID
                                                orderby order.ShipCountry, order.ShipCity, orderDetail.Product.ProductName, orderDetail.UnitPrice, orderDetail.Quantity
                                                select new OrderInfo
                                                {
                                                    Country = order.ShipCountry,
                                                    City = order.ShipCity,
                                                    Product = orderDetail.Product.ProductName,
                                                    Price = orderDetail.UnitPrice,
                                                    Quantity = orderDetail.Quantity,
                                                };

            _ordersData = ordersQuery.ToList();

            // customize outline styles
            _nodeStyle = new CellStyle[2];
            var cs = _nodeStyle[0] = _flex.Styles[CellStyleEnum.Subtotal0];
            cs.BackColor = Color.FromArgb(0xd0, 0xd0, 0xd0);
            cs.ForeColor = Color.Black;
            cs.Font = new Font(_flex.Font, FontStyle.Bold);
            cs = _nodeStyle[1] = _flex.Styles[CellStyleEnum.Subtotal1];
            cs.BackColor = Color.FromArgb(0xe0, 0xe0, 0xe0);
            cs.ForeColor = Color.Black;
            cs.Font = new Font(_flex.Font, FontStyle.Italic);

            // remove vertical gridlines from detail area
            _flex.Styles.Normal.Border.Direction = BorderDirEnum.Horizontal;

            ResetBinding();
        }

        #endregion

        //------------------------------------------------------------------------------------
        #region ** low-level methods: grouping, trees, subtotals with InsertNode and Aggregate

        // unbind and re-bind grid in order to reset everything
        void ResetBinding()
        {
            // re-bind grid
            _flex.DataSource = null;
            _flex.Cols.Fixed = fixedColumnToolStripMenuItem.Checked ? 1 : 0;
            _flex.Tree.Column = -1;
            _flex.DataSource = _ordersData;

            // format Price column
            _flex.Cols["Price"].Format = "n2";

            // format Cost column
            _flex.Cols["Cost"].Format = "n2";

            // resize columns to fit their content
            _flex.AutoSizeCols();

            // allow node content to spill onto next cell
            _flex.AllowMerging = AllowMergingEnum.Nodes;
        }

        // group on a given column inserting nodes of a given level
        void GroupBy(string columnName, int level)
        {
            object current = null;
            for (int r = _flex.Rows.Fixed; r < _flex.Rows.Count; r++)
            {
                if (!_flex.Rows[r].IsNode)
                {
                    var value = _flex[r, columnName];
                    if (!object.Equals(value, current))
                    {
                        // value changed: insert node, apply style
                        _flex.Rows.InsertNode(r, level);
                        _flex.Rows[r].Style = _nodeStyle[Math.Min(level, _nodeStyle.Length - 1)];

                        // show group name in first scrollable column
                        _flex[r, _flex.Cols.Fixed] = value;

                        // update current value
                        current = value;
                    }
                }
            }
        }

        // add subtotals to each node at a given level
        void AddSubtotals(int level, string colName)
        {
            // get column we are going to total on
            int colIndex = _flex.Cols.IndexOf(colName);

            // scan rows looking for nodes at the right level
            for (int r = _flex.Rows.Fixed; r < _flex.Rows.Count; r++)
            {
                if (_flex.Rows[r].IsNode)
                {
                    var node = _flex.Rows[r].Node;
                    if (node.Level == level)
                    {
                        // found a node, calculate the sum of cost
                        var range = node.GetCellRange();
                        var sum = _flex.Aggregate(AggregateEnum.Sum,
                            range.r1, colIndex, range.r2, colIndex,
                            AggregateFlags.ExcludeNodes);

                        // show the sum on the grid 
                        // (will use the column format automatically)
                        _flex[r, colIndex] = sum;
                    }
                }
            }
        }

        #endregion

        //------------------------------------------------------------------------------------
        #region ** event handlers

        void _btnReset_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetBinding();
            }
        }
        void _btnGroupCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetBinding();
                GroupBy("Country", 0);
            }
        }
        void _btnGroupCountryCity_Click(object sender, EventArgs e)
        {
            // suspend redrawing while updating
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group by Country, City
                GroupBy("Country", 0);
                GroupBy("City", 1);
            }
        }
        void _btnTreeCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // group by country as before
                _btnGroupCountry_Click(this, EventArgs.Empty);

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(0);
            }
        }
        void _btnTreeCountryCity_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // group by country and city as before
                _btnGroupCountryCity_Click(this, EventArgs.Empty);

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(1);
            }
        }
        void _btnNiceTreeCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // start with regular tree
                _btnTreeCountry_Click(this, EventArgs.Empty);

                // add totals per country
                AddSubtotals(0, "Cost");  // cost per country (level 0)
                AddSubtotals(0, "Quantity");       // quantity per country (level 0)

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
            }
        }
        void _btnNiceTreeCountryCity_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // start with regular tree
                _btnTreeCountryCity_Click(this, EventArgs.Empty);

                // add totals per country, city
                AddSubtotals(0, "Cost");  // cost per country (level 0)
                AddSubtotals(0, "Quantity");       // quantity per country (level 0)
                AddSubtotals(1, "Cost");  // cost per city (level 1)
                AddSubtotals(1, "Quantity");       // quantity per city (level 1)

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
                _flex.Cols["City"].Width = 0;
            }
        }

        void _btnGroupCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetBinding();
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "Cost");
            }
        }
        void _btnGroupCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group by Country, City
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "Cost");
                _flex.Subtotal(AggregateEnum.None, 1, "City", "Cost");
            }
        }
        void _btnTreeCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group by country as before
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "Cost");

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(0);
            }
        }
        void _btnTreeCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group by country and city as before
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "Cost");
                _flex.Subtotal(AggregateEnum.None, 1, "City", "Cost");

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(1);
            }
        }
        void _btnNiceTreeCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group and total by country
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Cost");
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Cost");

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(1);

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
            }
        }
        void _btnNiceTreeCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                _flex.Tree.Column = 0;

                // group and total by country and city
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Cost");
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Quantity");
                _flex.Subtotal(AggregateEnum.Sum, 1, "City", "Cost");
                _flex.Subtotal(AggregateEnum.Sum, 1, "City", "Quantity");

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
                _flex.Cols["City"].Width = 0;
                _flex.AllowMerging = AllowMergingEnum.Nodes;

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(1);
            }
        }

        void fixedColumnToolStripMenuItem_Click(object sender, EventArgs e)
        {
            fixedColumnToolStripMenuItem.Checked = !fixedColumnToolStripMenuItem.Checked;
            ResetBinding();
        }

        #endregion
    }

    /// <summary>
    /// Utility class used to encapsulate grid lengthy operations in BeginUpdate/EndUpdate block.
    /// This avoids flicker and ensures that all updates are done in case 
    /// an exception is thrown during the operation.
    /// </summary>
    class DeferRefresh : IDisposable
    {
        C1FlexGrid _grid;
        public DeferRefresh(C1FlexGrid grid)
        {
            _grid = grid;
            _grid.BeginUpdate();
        }
        public void Dispose()
        {
            _grid.EndUpdate();
        }
    }
}
